from ms_deisotope.utils import _MappingOverAttributeProxy

from .cv import Term, TermSet


class ActivationInformation(object):
    """Describes the dissociation process used to produce an MSn scan.

    Attributes
    ----------
    data : :class:`dict`
        Additional metadata describing the dissociation
    energy : float
        The quantity of energy used
    method : :class:`DissociationMethod`
        The dissociation method used
    """
    __slots__ = ('energy', 'method', 'data')

    def __init__(self, method, energy, data=None):  # pylint: disable=redefined-outer-name
        if data is None:
            data = dict()
        self.method = dissociation_methods_map.get(str(method).lower(), None)
        if self.method is None:
            try:
                self.method = dissociation_methods_map.get(
                    method.accession, None)
            except AttributeError:
                pass
            if self.method is None:
                self.method = self._make_unknown_method(method)
        self.energy = energy
        self.data = data

    def to_dict(self):
        d = {
            "method": str(self.method),
            "energy": self.energy,
            "data": self.data
        }
        return d

    @classmethod
    def from_dict(cls, d):
        if 'methods' in d and 'energies' in d:
            return MultipleActivationInformation(**d)
        else:
            return ActivationInformation(**d)

    @staticmethod
    def _make_unknown_method(method):  # pylint: disable=redefined-outer-name
        return DissociationMethod(UnknownDissociation.name, method, method, method, [method])

    def __repr__(self):
        return "ActivationInformation(%r, %r%s)" % (
            str(self.method), self.energy,
            "" if not self.data else ", %r" % self.data)

    def __str__(self):
        return str(self.method)

    def __hash__(self):
        return hash(str(self))

    def is_multiple_dissociation(self):
        """Determine if multiple dissociation methods were
        used.

        Returns
        -------
        bool
        """
        return False

    def has_supplemental_dissociation(self):
        """Determine if supplemental activation was used

        Returns
        -------
        bool
        """
        return False

    def __eq__(self, other):
        if other is None:
            return False
        if self.is_multiple_dissociation() != other.is_multiple_dissociation():
            return False
        if self.energy is not None:
            if other.energy is not None:
                energy_match = abs(self.energy - other.energy) < 1e-3
            else:
                energy_match = False
        else:
            if other.energy is not None:
                energy_match = False
            else:
                energy_match = True
        return (self.method == other.method) and energy_match

    def __ne__(self, other):
        return not self == other

    def has_dissociation_type(self, dissociation_type):
        return self.method.is_a(dissociation_type)

    @property
    def __dict__(self):
        return _MappingOverAttributeProxy(self)

    def __reduce__(self):
        return self.__class__, (self.method, self.energy, self.data)


class MultipleActivationInformation(ActivationInformation):
    """Describes the dissociation processes used to produce an MSn
    scan with multiple types of activation.

    Attributes
    ----------
    data : :class:`dict`
        Additional metadata describing the dissociation
    energies : :class:`list` of float
        A list of dissociation energies used. Parallel to :attr:`methods`
    methods : :class:`list` of :class:`DissociationMethod`
        A list of dissociation methods used. Parallel to :attr:`energies`
    """

    def __init__(self, methods, energies, data=None):  # pylint: disable=super-init-not-called
        if data is None:
            data = dict()
        self.methods = []
        for method in methods:  # pylint: disable=redefined-outer-name
            self.methods.append(
                dissociation_methods_map.get(
                    str(method).lower(), method))
        self.energies = list(energies)
        self.data = data

    def to_dict(self):
        d = {
            "methods": [str(m) for m in self.methods],
            "energies": self.energies,
            "data": self.data
        }
        return d

    @property
    def method(self):
        return self.methods[0]

    @property
    def energy(self):
        return self.energies[0]

    def __iter__(self):
        return iter(self.methods)

    def __repr__(self):
        return "MultipleActivationInformation(methods=%r, energies=%r%s)" % (
            list(map(str, self.methods)), self.energies,
            "" if not self.data else ", %r" % self.data)

    def __str__(self):
        return ', '.join(map(str, self.methods))

    def is_multiple_dissociation(self):
        return True

    def has_supplemental_dissociation(self):
        return any([d.is_supplemental() for d in self.methods])

    def __eq__(self, other):
        if other is None:
            return False
        if self.is_multiple_dissociation() != other.is_multiple_dissociation():
            return False
        if len(self.energies) != len(other.energies):
            return False
        energies_equal = []
        for se, oe in zip(self.energies, other.energies):
            if se is None and oe is None:
                energies_equal.append(True)
            elif se is None or oe is None:
                energies_equal.append(False)
            else:
                energies_equal.append(abs(se - oe) < 1e-3)
        return (self.methods == other.methods) and all(energies_equal)

    def has_dissociation_type(self, dissociation_type):
        return any(method.is_a(dissociation_type) for method in self.methods)

    def __reduce__(self):
        return self.__class__, (self.methods, self.energies, self.data)


class DissociationMethod(Term):
    """Controlled vocabulary describing a dissociation process
    """

    def is_supplemental(self):
        """Returns whether or not the dissociation process is
        supplemental

        Returns
        -------
        bool
        """
        return "supplemental" in self.name


dissociation_methods = []


# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('dissociation method', term_cls_name="DissociationMethod", writer=cog.out)
# ]]]
# CV Version: 4.1.95
dissociation_methods = TermSet([
    DissociationMethod('collision-induced dissociation', 'MS:1000133',
                       ('The dissociation of an ion after collisional excitation. The '
                        'term collisional-activated dissociation is not recommended.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('plasma desorption', 'MS:1000134',
                       ('The ionization of material in a solid sample by bombarding '
                        'it with ionic or neutral atoms formed as a result of the '
                        'fission of a suitable nuclide, typically 252Cf. Synonymous '
                        'with fission fragment ionization.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('post-source decay', 'MS:1000135',
                       ('A technique specific to reflectron time-of-flight mass '
                        'spectrometers where product ions of metastable transitions '
                        'or collision-induced dissociations generated in the drift '
                        'tube prior to entering the reflectron are m/z separated to '
                        'yield product ion spectra.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('surface-induced dissociation', 'MS:1000136',
                       ('Fragmentation that results from the collision of an ion with '
                        'a surface.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('blackbody infrared radiative dissociation', 'MS:1000242',
                       ('A special case of infrared multiphoton dissociation wherein '
                        'excitation of the reactant ion is caused by absorption of '
                        'infrared photons radiating from heated blackbody '
                        'surroundings, which are usually the walls of a vacuum '
                        'chamber. See also infrared multiphoton dissociation.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('electron capture dissociation', 'MS:1000250',
                       ('A process in which a multiply protonated molecules interacts '
                        'with a low energy electrons. Capture of the electron leads '
                        'the liberation of energy and a reduction in charge state of '
                        'the ion with the production of the (M + nH) (n-1)+ odd '
                        'electron ion, which readily fragments.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('sustained off-resonance irradiation', 'MS:1000282',
                       ('A technique associated with Fourier transform ion cyclotron '
                        'resonance (FT-ICR) mass spectrometry to carry out '
                        'ion/neutral reactions such as low-energy collision-induced '
                        'dissociation. A radio-frequency electric field of slightly '
                        'off-resonance to the cyclotron frequency of the reactant ion '
                        'cyclically accelerates and decelerates the reactant ion that '
                        "is confined in the Penning ion trap. The ion's orbit does "
                        'not exceed the dimensions of ion trap while the ion '
                        'undergoes an ion/neutral species process that produces a '
                        'high average translational energy for an extended time.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('low-energy collision-induced dissociation', 'MS:1000433',
                       ('A collision-induced dissociation process wherein the '
                        'precursor ion has the translational energy lower than '
                        'approximately 1000 eV. This process typically requires '
                        'multiple collisions and the collisional excitation is '
                        'cumulative.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('photodissociation', 'MS:1000435',
                       ('A process wherein the reactant ion is dissociated as a '
                        'result of absorption of one or more photons.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('electron transfer dissociation', 'MS:1000598',
                       ('A process to fragment ions in a mass spectrometer by '
                        'inducing fragmentation of cations (e.g. peptides or '
                        'proteins) by transferring electrons from radical-anions." '
                        '[DOI:10.1073/pnas.0402700101, PMID:15210983'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('pulsed q dissociation', 'MS:1000599',
                       ('A process that involves precursor ion activation at high Q, '
                        'a time delay to allow the precursor to fragment, then a '
                        'rapid pulse to low Q where all fragment ions are trapped. '
                        'The product ions can then be scanned out of the ion trap and '
                        'detected.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('in-source collision-induced dissociation', 'MS:1001880',
                       ('The dissociation of an ion as a result of collisional '
                        'excitation during ion transfer from an atmospheric pressure '
                        'ion source and the mass spectrometer vacuum.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('LIFT', 'MS:1002000',
                       ("A Bruker's proprietary technique where molecular ions are "
                        'initially accelerated at lower energy, then collide with '
                        "inert gas in a collision cell that is then 'lifted' to high "
                        'potential. The use of inert gas is optional, as it could '
                        'lift also fragments provided by LID." '
                        '[DOI:10.1007/s00216-003-2057-0'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('combined dissociation method', 'MS:1003181',
                       ('Combination of two or more dissociation methods that are '
                        'known by a special term.'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('negative electron transfer dissociation', 'MS:1003247',
                       ('A process to fragment ions in a mass spectrometer by '
                        'inducing fragmentation of anions (e.g. peptides or proteins) '
                        'by transferring electrons to a radical-cation." '
                        '[DOI:10.1016/j.jasms.2005.01.015'),
                       'dissociation method',
                       ['dissociation method']),
    DissociationMethod('beam-type collision-induced dissociation', 'MS:1000422',
                       ('A collision-induced dissociation process that occurs in a '
                        'beam-type collision cell.'),
                       'dissociation method',
                       ['collision-induced dissociation', 'dissociation method']),
    DissociationMethod('trap-type collision-induced dissociation', 'MS:1002472',
                       ('A collision-induced dissociation process that occurs in a '
                        'trap-type collision cell.'),
                       'dissociation method',
                       ['collision-induced dissociation', 'dissociation method']),
    DissociationMethod('supplemental collision-induced dissociation', 'MS:1002679',
                       ('The dissociation of an ion after supplemental collisional '
                        'excitation.'),
                       'dissociation method',
                       ['collision-induced dissociation', 'dissociation method']),
    DissociationMethod('infrared multiphoton dissociation', 'MS:1000262',
                       ('Multiphoton ionization where the reactant ion dissociates as '
                        'a result of the absorption of multiple infrared photons.'),
                       'dissociation method',
                       ['photodissociation', 'dissociation method']),
    DissociationMethod('ultraviolet photodissociation', 'MS:1003246',
                       ('Multiphoton ionization where the reactant ion dissociates as '
                        'a result of the absorption of multiple UV photons.'),
                       'dissociation method',
                       ['photodissociation', 'dissociation method']),
    DissociationMethod('electron-transfer/higher-energy collision dissociation', 'MS:1002631',
                       ('Dissociation process combining electron-transfer '
                        'dissociation and higher-energy collision dissociation. It '
                        'combines ETD (reaction time) followed by HCD (activation '
                        'energy).'),
                       'dissociation method',
                       ['combined dissociation method', 'dissociation method']),
    DissociationMethod('electron-transfer/collision-induced dissociation', 'MS:1003182',
                       ('Dissociation process combining electron-transfer '
                        'dissociation (ETD) and collision-induced dissociation (CID).'),
                       'dissociation method',
                       ['combined dissociation method', 'dissociation method']),
    DissociationMethod('higher energy beam-type collision-induced dissociation', 'MS:1002481',
                       ('A collision-induced dissociation process wherein the '
                        'projectile ion has the translational energy higher than '
                        'approximately 1000 eV.'),
                       'dissociation method',
                       ['beam-type collision-induced dissociation', 'collision-induced dissociation', 'dissociation method']),
    DissociationMethod('supplemental beam-type collision-induced dissociation', 'MS:1002678',
                       ('A supplemental collision-induced dissociation process that '
                        'occurs in a beam-type collision cell in addition to another '
                        'primary type of dissociation.'),
                       'dissociation method',
                       ['beam-type collision-induced dissociation', 'collision-induced dissociation', 'dissociation method']),
])
# [[[end]]]


UnknownDissociation = DissociationMethod(
    "unknown dissociation", None, None, 'dissociation method',
    [u'dissociation method'])


dissociation_methods_map = {
    "": UnknownDissociation,
    None: UnknownDissociation,
    UnknownDissociation.name: UnknownDissociation
}


method = None
for method in dissociation_methods:
    dissociation_methods_map[method.name] = method
    dissociation_methods_map[method.id] = method
del method


CID = dissociation_methods_map.get("collision-induced dissociation")
HCD = dissociation_methods_map.get("beam-type collision-induced dissociation")
ETD = dissociation_methods_map.get("electron transfer dissociation")
ECD = dissociation_methods_map.get("electron capture dissociation")

supplemental_term_map = {
    dissociation_methods_map["beam-type collision-induced dissociation"]: dissociation_methods_map[
        "supplemental beam-type collision-induced dissociation"],
    dissociation_methods_map["collision-induced dissociation"]: dissociation_methods_map[
        "supplemental collision-induced dissociation"]
}


supplemental_energy = "supplemental collision energy"


energy_terms = set([
    "collision energy",
    supplemental_energy,
    "activation energy",
    "collision energy ramp start",
    "collision energy ramp end",
    "percent collision energy ramp start",
    "percent collision energy ramp end",
])


dissociation_methods_map.update({
    'cad': CID,
    'cid': CID,
    "hcd": HCD,
    "etd": ETD,
    "ecd": ECD
})


ActivationInformation.dissociation_methods = dissociation_methods_map


__all__ = [
    "ActivationInformation", "MultipleActivationInformation", "DissociationMethod",
    "dissociation_methods", "UnknownDissociation", "dissociation_methods_map",
    "CID", "HCD", "ETD", "ECD", "supplemental_energy", "supplemental_term_map",
    "energy_terms"
]



from ms_deisotope._c.feature_map.lcms_feature cimport LCMSFeature
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM


cdef class LCMSFeatureMap(object):
    cdef:
        public list features

    cdef Py_ssize_t get_size(self)
    cdef LCMSFeature get(self, size_t i)
    cpdef list find_all(self, double mz, double error_tolerance=*)
    cpdef LCMSFeature search(self, double mz, double error_tolerance=*)
    cpdef list between(self, double lo, double hi, double error_tolerance=*)
    cpdef list spanning_time(self, double time_point)
    cpdef LCMSFeatureMap clone(self, bint deep=*)

cpdef tuple binary_search_with_flag(list array, double mz, double error_tolerance)
cdef long binary_search(list array, double mz, double error_tolerance)
cdef void search_sweep(list array, double mz, double error_tolerance, long* loout, long* hiout)

cpdef list split_sparse(LCMSFeatureMap self, double delta_rt=*, size_t min_size=*)



from .scan_clustering import (
    cluster_scans, SpectrumCluster, SpectrumClusterCollection,
    ScanClusterBuilder, iterative_clustering,
    ScanClusterWriter, ScanClusterReader,
    JSONScanClusterWriter, JSONScanClusterReader)
from .similarity_methods import peak_set_similarity

__all__ = [
    "cluster_scans", "peak_set_similarity",
    "SpectrumCluster"
]



import sys
import traceback
import os
import platform

from setuptools import setup, Extension as _Extension, find_packages

from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)


def has_option(name):
    try:
        sys.argv.remove('--%s' % name)
        return True
    except ValueError:
        pass
    # allow passing all cmd line options also as environment variables
    env_val = os.getenv(name.upper().replace('-', '_'), 'false').lower()
    if env_val == "true":
        return True
    return False


include_diagnostics = has_option("include-diagnostics")
force_cythonize = has_option("force-cythonize")
no_openmp = has_option('no-openmp')
debug_symbols = has_option("debug")


with_openmp = not no_openmp


def configure_openmp(ext):
    # http://www.microsoft.com/en-us/download/confirmation.aspx?id=2092 was required.
    if os.name == 'nt' and with_openmp:
        ext.extra_compile_args.append("/openmp")
    elif platform.system() == 'Darwin':
        pass
    elif with_openmp:
        ext.extra_compile_args.append("-fopenmp")
        ext.extra_link_args.append("-fopenmp")


def Extension(*args, **kwargs):
    ext = _Extension(*args, **kwargs)
    if debug_symbols:
        if os.name == 'nt':
            ext.extra_compile_args.append("-Zi")
            ext.extra_compile_args.append("-Ox")
            ext.extra_link_args.append("-debug:full")
        else:
            ext.extra_compile_args.append("-g3")
            ext.extra_compile_args.append("-O0")
    # ext.extra_compile_args.append("/fsanitize=address")
    return ext

def OpenMPExtension(*args, **kwargs):
    ext = Extension(*args, **kwargs)
    configure_openmp(ext)
    return ext


def make_extensions():
    is_ci = bool(os.getenv("CI", ""))
    try:
        import numpy
    except ImportError:
        print("Installation requires `numpy`")
        raise
    try:
        import brainpy
    except ImportError:
        print("Installation requires `brainpy`, install with `python -m pip install brain-isotopic-distribution`")
        raise
    try:
        import ms_peak_picker
    except ImportError:
        print("Installation requires `ms_peak_picker`")
        raise
    macros = []
    try:
        from Cython.Build import cythonize
        cython_directives = {
            'embedsignature': True,
            "profile": include_diagnostics
        }
        if include_diagnostics:
            macros.append(("CYTHON_TRACE_NOGIL", "1"))
        if is_ci and include_diagnostics:
            cython_directives['linetrace'] = True
        extensions = cythonize([
            Extension(name='ms_deisotope._c.scoring', sources=["src/ms_deisotope/_c/scoring.pyx"],
                      include_dirs=[brainpy.get_include(), ms_peak_picker.get_include(), numpy.get_include()]),
            Extension(name='ms_deisotope._c.averagine', sources=["src/ms_deisotope/_c/averagine.pyx"],
                      include_dirs=[brainpy.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.deconvoluter_base', sources=["src/ms_deisotope/_c/deconvoluter_base.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.peak_set', sources=["src/ms_deisotope/_c/peak_set.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()]),
            Extension(name='ms_deisotope._c.feature_map.lcms_feature', sources=[
                      "src/ms_deisotope/_c/feature_map/lcms_feature.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.feature_processor', sources=[
                      "src/ms_deisotope/_c/feature_map/feature_processor.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.feature_map', sources=[
                      "src/ms_deisotope/_c/feature_map/feature_map.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.feature_fit', sources=[
                      "src/ms_deisotope/_c/feature_map/feature_fit.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.shape_fitter', sources=[
                      "src/ms_deisotope/_c/feature_map/shape_fitter.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.profile_transform', sources=[
                      "src/ms_deisotope/_c/feature_map/profile_transform.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.dependence_network', sources=[
                      "src/ms_deisotope/_c/feature_map/dependence_network.pyx"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.utils', sources=["src/ms_deisotope/_c/utils.pyx"],
                      include_dirs=[brainpy.get_include(), ms_peak_picker.get_include(), numpy.get_include()]),
            Extension(name='ms_deisotope._c.peak_dependency_network.intervals',
                      sources=['src/ms_deisotope/_c/peak_dependency_network/intervals.pyx'],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.peak_dependency_network.subgraph',
                      sources=['src/ms_deisotope/_c/peak_dependency_network/subgraph.pyx'],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.peak_dependency_network.peak_network',
                      sources=['src/ms_deisotope/_c/peak_dependency_network/peak_network.pyx'],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.spectrum_graph',
                      sources=['src/ms_deisotope/_c/spectrum_graph.pyx'],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.feature_graph',
                      sources=['src/ms_deisotope/_c/feature_map/feature_graph.pyx'],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            OpenMPExtension(
                name='ms_deisotope._c.similarity_methods', sources=["src/ms_deisotope/_c/similarity_methods.pyx"],
                include_dirs=[brainpy.get_include(), ms_peak_picker.get_include(), numpy.get_include()]),
            Extension(
                name='ms_deisotope._c.units', sources=["src/ms_deisotope/_c/units.pyx"],
                include_dirs=[]),
        ],
        compiler_directives=cython_directives,
        emit_linenums=True,
        force=force_cythonize)
    except ImportError:
        extensions = ([
            Extension(name='ms_deisotope._c.scoring', sources=["src/ms_deisotope/_c/scoring.c"],
                      include_dirs=[brainpy.get_include(), ms_peak_picker.get_include(), numpy.get_include()]),
            Extension(name='ms_deisotope._c.averagine', sources=["src/ms_deisotope/_c/averagine.c"],
                      include_dirs=[brainpy.get_include()]),
            Extension(name='ms_deisotope._c.deconvoluter_base', sources=["src/ms_deisotope/_c/deconvoluter_base.c"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()]),
            Extension(name='ms_deisotope._c.peak_set', sources=["src/ms_deisotope/_c/peak_set.c"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()]),
            Extension(name='ms_deisotope._c.feature_map.lcms_feature', sources=[
                      "src/ms_deisotope/_c/feature_map/lcms_feature.c"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()]),
            Extension(name='ms_deisotope._c.feature_map.feature_processor', sources=[
                      "src/ms_deisotope/_c/feature_map/feature_processor.c"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()]),
            Extension(name='ms_deisotope._c.feature_map.feature_map', sources=[
                      "src/ms_deisotope/_c/feature_map/feature_map.c"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()]),
            Extension(name='ms_deisotope._c.feature_map.feature_fit', sources=[
                      "src/ms_deisotope/_c/feature_map/feature_fit.c"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()]),
            Extension(name='ms_deisotope._c.feature_map.shape_fitter', sources=[
                      "src/ms_deisotope/_c/feature_map/shape_fitter.c"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.profile_transform', sources=[
                      "src/ms_deisotope/_c/feature_map/profile_transform.c"],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include(), brainpy.get_include()]),
            Extension(name='ms_deisotope._c.feature_map.dependence_network', sources=[
                      "src/ms_deisotope/_c/feature_map/dependence_network.c"],
                      include_dirs=[
                          numpy.get_include(), ms_peak_picker.get_include()],
                      define_macros=macros),
            Extension(name='ms_deisotope._c.utils', sources=["src/ms_deisotope/_c/utils.c"],
                      include_dirs=[brainpy.get_include(), ms_peak_picker.get_include(), numpy.get_include()]),
            Extension(name='ms_deisotope._c.peak_dependency_network.intervals',
                      sources=['src/ms_deisotope/_c/peak_dependency_network/intervals.c'],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.peak_dependency_network.subgraph',
                      sources=['src/ms_deisotope/_c/peak_dependency_network/subgraph.c'],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.peak_dependency_network.peak_network',
                      sources=['src/ms_deisotope/_c/peak_dependency_network/peak_network.c'],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.spectrum_graph',
                      sources=['src/ms_deisotope/_c/spectrum_graph.c'],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            Extension(name='ms_deisotope._c.feature_map.feature_graph',
                      sources=['src/ms_deisotope/_c/feature_map/feature_graph.c'],
                      include_dirs=[numpy.get_include(), ms_peak_picker.get_include()], define_macros=macros),
            OpenMPExtension(
                name='ms_deisotope._c.similarity_methods', sources=["src/ms_deisotope/_c/similarity_methods.c"],
                include_dirs=[brainpy.get_include(), ms_peak_picker.get_include(), numpy.get_include()]),
            Extension(
                name='ms_deisotope._c.units', sources=["src/ms_deisotope/_c/units.c"],
                include_dirs=[]),
        ])
    return extensions


ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32':
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    ext_errors += (IOError,)


class BuildFailed(Exception):

    def __init__(self):
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax

    def __str__(self):
        return str(self.cause)


class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            traceback.print_exc()
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors:
            traceback.print_exc()
            raise BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            traceback.print_exc()
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise BuildFailed()
            raise


cmdclass = {}

cmdclass['build_ext'] = ve_build_ext


def status_msgs(*msgs):
    print('*' * 75)
    for msg in msgs:
        print(msg)
    print('*' * 75)


install_requires = [
    "numpy",
    "scipy",
    "six",
    "dill",
    "ms_peak_picker >= 0.1.38",
    "brain-isotopic-distribution >= 1.5.8",
    "pyteomics >= 4.5",
    "lxml",
    "psims >= 0.1.44",
    "python-idzip >=0.3.2",
    "pyzstd",
]


extra_requires = {
    "com": [
        "comtypes"
    ],
    "net": [
        "pythonnet"
    ],
    "plot": [
        "matplotlib",
    ],
    "cli": [
        'click'
    ],
    'mzmlb': [
        'h5py',
        'hdf5plugin'
    ]
}


extra_requires['all'] = [dep for feature_reqs in extra_requires.values() for dep in feature_reqs]


def run_setup(include_cext=True):
    with open("src/ms_deisotope/version.py") as version_file:
        version = None
        for line in version_file.readlines():
            if "version = " in line:
                version = line.split(" = ")[1].replace("\"", "").strip()
                print("Version is: %r" % (version,))
                break
        else:
            print("Cannot determine version")

    try:
        with open("README.rst") as readme_file:
            long_description = readme_file.read()
    except Exception as e:
        print(e)
        long_description = ''

    setup(
        name='ms_deisotope',
        version=version,
        packages=find_packages(where='src'),
        package_dir={"": "src"},
        author=', '.join(["Joshua Klein"]),
        author_email=','.join(["jaklein@bu.edu"]),
        description='Access, Deisotope, and Charge Deconvolute Mass Spectra',
        long_description=long_description,
        ext_modules=make_extensions() if include_cext else None,
        cmdclass=cmdclass,
        entry_points={
            'console_scripts': [
                "ms-index = ms_deisotope.tools.indexing:main",
                "ms-view = ms_deisotope.tools.view:main",
                "ms-deisotope = ms_deisotope.tools.deisotoper.main:deisotope",
            ],
        },
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Topic :: Scientific/Engineering :: Bio-Informatics'],
        install_requires=install_requires,
        extras_require=extra_requires,
        include_package_data=True,
        zip_safe=False,
        python_requires=">3.8",
        project_urls={
            'Documentation': 'https://mobiusklein.github.io/ms_deisotope',
            'Source Code': 'https://github.com/mobiusklein/ms_deisotope',
            'Issue Tracker': 'https://github.com/mobiusklein/ms_deisotope/issues',
            'Change Log': 'https://github.com/mobiusklein/ms_deisotope/blob/master/CHANGELOG.md'
        })


try:
    run_setup(True)
except Exception as exc:
    print(exc)
    traceback.print_exc()
    run_setup(False)

    status_msgs(
        "WARNING: The C extension could not be compiled, " +
        "speedups are not enabled.",
        "Plain-Python build succeeded."
    )



"""Implements the infrastructure to spread indexing tasks over a single
:class:`~.RandomAccessScanSource` across multiple processes for speed.

The primary function in this module is :func:`index`, which will perform
this dispatch.
"""
import multiprocessing
import logging
from typing import Any, Callable, Dict, Iterator, List, Tuple, Union, Optional

import dill
from ms_deisotope.data_source.scan.base import ScanBunch

from ms_deisotope.data_source.scan.loader import RandomAccessScanSource, ScanIterator

from .scan_index import ExtendedScanIndex
from .scan_interval_tree import (
    ScanIntervalTree, extract_intervals, make_rt_tree)

logger = logging.getLogger(__name__)
n_cores = multiprocessing.cpu_count()


ScanGenerator = Union[ScanIterator, RandomAccessScanSource]
_IndexingCallable = Callable[[Tuple[ScanGenerator, int, int]], Tuple[float, Any]]


def indexing_iterator(reader: ScanGenerator, start: int, end: int, index: ExtendedScanIndex) -> Iterator[ScanBunch]:
    """A helper function which will iterate over an interval of a :class:`~.RandomAccessScanSource`
    while feeding each yielded :class:`~.ScanBunch` into a provided :class:`~.ExtendedScanIndex`.

    Parameters
    ----------
    reader : :class:`~.RandomAccessScanSource` or :class:`~.ScanIterator`
        The scan data source to loop over.
    start : int
        The starting index
    end : int
        The stopping index
    index : :class:`~.ExtendedScanIndex`
        The scan index to fill.

    Yields
    ------
    :class:`~.ScanBunch`
    """
    assert end >= start, "End cannot precede Start"
    try:
        iterator = reader.start_from_scan(index=start, grouped=True)
    except AttributeError:
        if start != 0:
            raise
        iterator = reader
    for scan_bunch in iterator:
        try:
            ix = scan_bunch.precursor.index
        except AttributeError:
            ix = scan_bunch.products[0].index
        if ix > end:
            break
        index.add_scan_bunch(scan_bunch)
        yield scan_bunch

def index_chunk(reader: ScanGenerator, start: int, end: int) -> Tuple[int, int, ExtendedScanIndex, List]:
    """The task function for :func:`quick_index`, which will build an
    :class:`~.ExtendedIndex` and :class:`ScanIntervalTree` from an index
    range over a :class:`~.ScanIterator`

    Parameters
    ----------
    reader : :class:`~.ScanIterator` or :class:`~.RandomAccessScanSource`
        The scan source to iterate over
    start : int
        The starting index
    end : int
        The stopping index

    Returns
    -------
    start: int
        The starting index for this chunk
    end: int
        The stopping index for this chunk
    index: :class:`~.ExtendedIndex`
        The constructed scan metadata index for this chunk
    intervals: :class:`~.ScanIntervalTree
        The constructed scan interval tree for this chunk
    """
    index = ExtendedScanIndex()
    iterator = indexing_iterator(reader, start, end, index)
    intervals = extract_intervals(iterator)
    return (start, end, index, intervals)


def partition_work(n_items: int, n_workers: int, start_index: int=0,
                   reader: Optional[ScanGenerator]=None) -> List[Tuple[int, int]]:
    """Given an index range and a number of workers to work on them,
    break the index range into approximately evenly sized sub-intervals.

    This is a helper function for :func:`run_task_in_chunks` used to
    compute the chunks.

    Parameters
    ----------
    n_items : int
        The maximum value of the index range
    n_workers : int
        The number of workers to split the work between
    start_index : int, optional
        The starting value of the index range (the default is 0)

    Returns
    -------
    list:
        A list of (start, end) pairs defining the index ranges each worker will
        handle.
    """
    if n_workers == 1:
        return [[start_index, start_index + n_items]]
    chunk_size = int(n_items / n_workers)
    n_items += start_index
    intervals = []
    start = start_index
    end = start + chunk_size
    if reader is not None:
        if reader.has_fast_random_access:
            scan = reader.get_scan_by_index(start)
            if scan.ms_level > 1:
                start_scan = reader.find_previous_ms1(start)
                if start_scan is not None:
                    start = start_scan.index
            end = min(end, len(reader) - 1)
            scan = scan = reader.get_scan_by_index(end)
            if scan.ms_level > 1:
                end_scan = reader.find_next_ms1(end)
                if end_scan is not None:
                    end = end_scan.index
    intervals.append([start, end])
    start += chunk_size

    while start  < n_items:
        end = start + chunk_size
        if reader is not None:
            scan = reader.get_scan_by_index(start)
            if scan.ms_level > 1:
                start_scan = reader.find_previous_ms1(start)
                if start_scan is not None:
                    start = start_scan.index
            end = min(end, len(reader))
            if end != len(reader):
                scan = scan = reader.get_scan_by_index(end)
                if scan.ms_level > 1:
                    end_scan = reader.find_next_ms1(end)
                    if end_scan is not None:
                        end = end_scan.index
        intervals.append([start, end])
        start = end
    # Make sure that the last chunk actually covers the end of
    # the interval.
    last = intervals[-1]
    if last[1] > n_items:
        last[1] = n_items
    else:
        last[1] += (n_items - last[1])
    return intervals


class _Indexer(object):
    """A pickle-able callable object which wraps :func:`index_chunk` for the call signature
    used by :func:`run_task_in_chunks`
    """
    def __call__(self, payload: Tuple[ScanIterator, float, float]):
        reader, start, end = payload
        try:
            result = index_chunk(reader, start, end)
        except Exception as e:
            print(reader, start, end, e)
            import traceback
            traceback.print_exc()
            raise e
        return result


class _TaskWrapper(object):
    """A simple wrapper for a callable to capture the index range for a chunk created by
    :func:`run_task_in_chunks`, so that the start index of the chunk is known.
    """

    task: _IndexingCallable

    def __init__(self, task):
        self.task = task

    def __getstate__(self):
        state = {
            "task": dill.dumps(self.task)
        }
        return state

    def __setstate__(self, state):
        self.task = dill.loads(state['task'])

    def __call__(self, payload) -> Tuple[float, Any]:
        _reader, start, _end = payload
        out = self.task(payload)
        return start, out


class _TaskPayload(object):
    """A wrapper for the input to the distributed task which transmits the :class:`~.RandomAccessScanSource`
    via :mod:`dill` to make pickling of any wrapped file objects possible.

    Mocks a subset of the :class:`~.Sequence` API to allow it to be treated like a :class:`tuple`

    Attributes
    ----------
    reader: :class:`~.RandomAccessScanSource`
        The scan data source to be shared with the worker.
    start: int
        The scan index to start processing from
    end: int
        The scan index to stop processing at.
    options: dict
        A dictionary of extra arguments that the task might use.
    """

    reader: ScanGenerator
    start: int
    end: int
    options: Dict[str, Any]

    def __init__(self, reader, start, end, **kwargs):
        self.reader = reader
        self.start = start
        self.end = end
        self.options = kwargs

    def __iter__(self):
        yield self.reader
        yield self.start
        yield self.end

    def __getitem__(self, i):
        if i == 0:
            return self.reader
        elif i == 1:
            return self.start
        elif i == 2:
            return self.end
        else:
            raise IndexError(i)

    def __len__(self):
        return 3

    def __getstate__(self):
        state = {
            "reader": dill.dumps(self.reader, -1),
            "start": self.start,
            "end": self.end,
            "options": self.options
        }
        return state

    def __setstate__(self, state):
        self.reader = dill.loads(state['reader'])
        self.start = state['start']
        self.end = state['end']
        self.options = state['options']

    def __repr__(self):
        template = "{self.__class__.__name__}({self.reader}, {self.start}, {self.end}, {self.options})"
        return template.format(self=self)


def run_task_in_chunks(reader: ScanGenerator,
                       n_processes: Optional[int]=None,
                       n_chunks: Optional[int]=None,
                       scan_interval: Tuple[float, float]=None,
                       task: Optional[_IndexingCallable] = None,
                       progress_indicator: Optional[Callable] = None):
    """Run a :class:`~.Callable` `task` over a :class:`~.ScanIterator` in chunks across multiple processes.

    This function breaks apart a :class:`~.ScanIterator`'s scans over `scan_interval`,
    or the whole sequence if not provided.

    Parameters
    ----------
    reader : :class:`~.ScanIterator`
        The set of :class:`~.Scan` objects to operate on
    n_processes : int, optional
        The number of worker processes to use (the default is 4 or the number of cores available,
        whichever is lower)
    n_chunks : int, optional
        The number of chunks to break the scan range into (the default is equal to `n_processes`)
    scan_interval : :class:`tuple` of (:class:`int`, :class:`int`), optional
        The start and stop scan index to apply the task over. If omitted, the entire scan range will
        be used. If either entry is :const:`None`, then the index will be assumed to be the first or
        last scan respectively.
    task : :class:`~.Callable`
        The callable object which will be executed on each chunk in a sub-process. It must take one
        argument, a :class:`tuple` of (`reader`, `start index`, `stop index`), and it must return a pickle-able
        object.
    progress_indicator : :class:`~.Callable`, optional
        A callable object which will be used to report progress as chunks finish processing. It must take
        one argument, a :class:`float` which represents the fraction of all work completed.

    Returns
    -------
    :class:`list`:
        The result of `task` on each chunk of `reader` in index sorted order.

    """
    if n_processes is None:
        n_processes = min(n_cores, 4)
    if task is None or not callable(task):
        raise ValueError("The task must be callable!")
    if scan_interval is None:
        start_scan = 0
        end_scan = len(reader.index)
    else:
        start_scan, end_scan = scan_interval
        if start_scan is None:
            start_scan = 0
        if end_scan is None:
            end_scan = len(reader.index)
    if n_chunks is None:
        n_chunks = n_processes
    n_items = end_scan - start_scan
    pool = multiprocessing.Pool(n_processes)
    scan_ranges = partition_work(
        n_items, n_chunks, start_scan, reader=reader)
    feeder = (
        _TaskPayload(reader, scan_range[0], scan_range[1])
        for scan_range in scan_ranges)
    result = []
    for i, block in enumerate(pool.imap_unordered(_TaskWrapper(task), feeder), 1):
        result.append(block)
        if progress_indicator is not None:
            progress_indicator(i / float(n_chunks))
    pool.close()
    pool.join()
    result.sort(key=lambda x: x[0])
    result = [x[1] for x in result]
    return result


def _merge_indices(indices: List[ExtendedScanIndex]) -> ExtendedScanIndex:
    index = indices[0]
    for ind in indices:
        index = index.merge(ind)
    return index


def _make_interval_tree(intervals) -> ScanIntervalTree:
    concat = []
    for i in intervals:
        concat.extend(i)
    return ScanIntervalTree(make_rt_tree(concat), None)


def index(reader: ScanGenerator, n_processes: int=4, scan_interval: Tuple[float, float]=None,
          progress_indicator: Optional[Callable]=None, chunks_per_worker: int=3) -> Tuple[ExtendedScanIndex, ScanIntervalTree]:
    """Generate a :class:`~.ExtendedScanIndex` and :class:`~.ScanIntervalTree` for
    `reader` between `scan_interval` start and end points across `n_processes` worker
    processes.

    If a :class:`~.ScanIterator` is passed instead of a :class:`~.RandomAccessScanSource`,
    only a single process will be used.

    Parameters
    ----------
    reader : :class:`~.RandomAccessScanSource` or :class:`~.ScanIterator`
        The scan data source to index.
    n_processes : int, optional
        The number of worker processes to use (the default is 4 or however many CPUs are available, whichever is lower)
    scan_interval : tuple, optional
        The start and stop scan indices to operate on (the default is None, which will index the entire file)
    progress_indicator : :class:`Callable`, optional
        A callable object which will be used to report progress as chunks finish processing. It must take
        one argument, a :class:`float` which represents the fraction of all work completed.

    Returns
    -------
    :class:`~.ExtendedScanIndex`:
        The extended metadata index for this data file.
    :class:`~.ScanIntervalTree`:
        The scan interval tree for this data file.

    See Also
    --------
    run_task_in_chunks
    """
    task = _Indexer()
    # indexing a :class:`ScanIterator` without random access, have to go in sequence
    if not hasattr(reader, 'start_from_scan'):
        logger.info("A non-random access ScanIterator was passed, defaulting to a single worker.")
        reader.make_iterator(grouped=True)
        chunks = [task(_TaskPayload(reader, 0, len(reader)))]
        n_processes = 1
    else:
        chunks = run_task_in_chunks(
            reader,
            n_processes,
            n_chunks=n_processes * chunks_per_worker,
            scan_interval=scan_interval,
            task=task,
            progress_indicator=progress_indicator)
    indices = [chunk[2] for chunk in chunks]
    intervals = [chunk[3] for chunk in chunks]
    index = _merge_indices(indices)
    interval_tree = _make_interval_tree(intervals)
    return index, interval_tree


def multi_index(readers, n_processes=4, scan_interval=None):
    index_collection = []
    interval_collection = []
    for reader in readers:
        chunks = run_task_in_chunks(
            reader, n_processes, scan_interval=scan_interval, task=_Indexer())
        indices = [chunk[2] for chunk in chunks]
        intervals = [chunk[3] for chunk in chunks]
        index_collection.append(indices)
        interval_collection.append(intervals)



# from .libload import proxy, register_dll, _register_dll
# from .loader import (MassLynxRawLoader, is_waters_raw_dir,
#                      determine_if_available, infer_reader,
#                      IndexEntry, Cycle)


# __all__ = [
#     "proxy", "register_dll", "_register_dll",
#     "MassLynxRawLoader", "is_waters_raw_dir",
#     "determine_if_available", "infer_reader",
#     "IndexEntry", "Cycle"
# ]


from ms_deisotope._c.peak_dependency_network.intervals cimport SpanningMixin, SimpleInterval

from ms_deisotope._c.feature_map.lcms_feature cimport LCMSFeature

cdef class FeatureGraphNode(SpanningMixin):
    cdef:
        public LCMSFeature feature
        public size_t index
        public set edges
        public double center
        public double mz
        public object generator # Factor this out somehow. Don't want to add a `__dict__`

    cpdef double _average_center(self)

cdef class FeatureGraphEdge(object):

    cdef:
        public FeatureGraphNode node_a
        public FeatureGraphNode node_b
        public object transition
        public double weight
        public double mass_error
        public double rt_error
        public Py_hash_t _hash

    cpdef remove(self)


cdef class DeconvolutedFeatureGraphNode(FeatureGraphNode):
    cdef:
        public double neutral_mass
        public int charge


cdef class IonMobilityProfileFeatureGraphNode(DeconvolutedFeatureGraphNode):
    cdef:
        public SimpleInterval ion_mobility_interval


cdef class TimeQuery(SpanningMixin):
    pass


cdef class PPMQuery(SpanningMixin):
    pass


cpdef list connected_components(self)


from . import peak_network, intervals, subgraph, utils

from .peak_network import (
    PeakNode, DependenceCluster, PeakDependenceGraph,
    NetworkedTargetedDeconvolutionResult, NoIsotopicClustersError)

from .subgraph import (
    ConnectedSubgraph, FitNode,
    GreedySubgraphSelection)

from .intervals import (
    Interval, IntervalTreeNode, SpanningMixin)

try:
    from .interval_viz import draw_envelope_subgraph
except ImportError:
    pass


__all__ = [
    "PeakNode",
    "DependenceCluster",
    "PeakDependenceGraph",
    "FitNode",
    "ConnectedSubgraph",
    "GreedySubgraphSelection",
    "Interval",
    "IntervalTreeNode",
    "SpanningMixin",
    "peak_network",
    "intervals",
    "subgraph",
    "utils",
    "draw_envelope_subgraph"
]



from ms_deisotope.constants import (ERROR_TOLERANCE, TRUNCATE_AFTER, IGNORE_BELOW)
from ms_deisotope.averagine import PROTON

from .utils import charge_range_, drop_placeholders
from .averagine_based import AveraginePeakDependenceGraphDeconvoluter
from .composition_list import CompositionListDeconvoluterBase

_APDGD = AveraginePeakDependenceGraphDeconvoluter


class HybridAveragineCompositionListPeakDependenceGraphDeconvoluter(_APDGD, CompositionListDeconvoluterBase):
    def __init__(self, peaklist, composition_list=None, *args, **kwargs):
        if composition_list is None:
            composition_list = []
        super(HybridAveragineCompositionListPeakDependenceGraphDeconvoluter, self).__init__(
            peaklist, composition_list=composition_list, *args, **kwargs)

    def deconvolute_composition(self, composition, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                                charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW,
                                mass_shift=None):
        for charge in charge_range_(*charge_range):
            fit = self.fit_composition_at_charge(
                composition, charge, error_tolerance, charge_carrier=charge_carrier,
                truncate_after=truncate_after, mass_shift=mass_shift, ignore_below=ignore_below)
            if fit is None:
                continue
            rep_eid = drop_placeholders(fit.experimental)
            if len(rep_eid) == 1 and fit.charge > 1:
                continue
            if not self.scorer.reject(fit):
                fit.data = (composition, charge)
                self.peak_dependency_network.add_fit_dependence(fit)
            if self.incremental_truncation is not None:
                for case in self.fit_incremental_truncation(fit, self.incremental_truncation):
                    if not self.scorer.reject(case):
                        self.peak_dependency_network.add_fit_dependence(case)

    def populate_graph(self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), left_search_limit=1,
                       right_search_limit=0, charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
                       ignore_below=IGNORE_BELOW, mass_shift=None):
        for composition in self.composition_list:
            self.deconvolute_composition(
                composition, error_tolerance, charge_range=charge_range,
                truncate_after=truncate_after, charge_carrier=charge_carrier,
                mass_shift=mass_shift)
        AveraginePeakDependenceGraphDeconvoluter.populate_graph(
            self, error_tolerance,
            charge_range=charge_range,
            left_search_limit=left_search_limit,
            right_search_limit=right_search_limit,
            charge_carrier=charge_carrier, truncate_after=truncate_after)



"""Thermo RAW file reading implementation using the Windows-specific COM
interface provided by MSFileReader.dll library provided by Thermo.

This module provides :class:`ThermoRawLoader`, a :class:`~.RandomAccessScanSource`
implementation.

Depends upon :mod:`comtypes` to interact with the operating system's
COM API and to register and create COM wrapper modules on demand.

The public interface of this module should be identical to
:mod:`ms_deisotope.data_source.thermo_raw_net`.

.. note::
    This interface was largely based upon the APIs that ProteoWizard used, both
    in order to understand how the Thermo libraries really worked, and to maintain
    parity with it.

"""
# pragma: no cover

import os
import warnings
from collections import OrderedDict

import logging

import numpy as np

from pyteomics.xml import unitfloat

from ms_deisotope.data_source.common import (
    ScanDataSource, RandomAccessScanSource,
    Scan, PrecursorInformation, ChargeNotProvided,
    ActivationInformation, IsolationWindow,
    ScanAcquisitionInformation, ScanEventInformation, ScanWindow,
    MultipleActivationInformation)

from ms_deisotope.data_source._thermo_helper import (
    _RawFileMetadataLoader, analyzer_map,
    FilterString, _id_template, _InstrumentMethod,
    _make_id, ThermoRawScanPtr)

from ms_deisotope.data_source.metadata.activation import (
    supplemental_term_map, dissociation_methods_map)
from ms_deisotope.data_source.metadata.sample import Sample
from ms_deisotope.data_source.metadata.scan_traits import FAIMS_compensation_voltage

try:
    from ms_deisotope.data_source._vendor.MSFileReader import (  # pylint: disable=unused-import
        ThermoRawfile as _ThermoRawFileAPI, register_dll,
        log as _api_logger, safearray_as_ndarray)

    comtypes_logger = logging.getLogger("comtypes")
    comtypes_logger.setLevel("INFO")
    _api_logger.setLevel("INFO")

    def is_thermo_raw_file(path):
        """Detect whether or not the file referenced by ``path``
        is a Thermo RAW file.

        Parameters
        ----------
        path: :class:`str`
            The path to test

        Returns
        -------
        :class:`bool`:
            Whether or not the file is a Thermo RAW file.
        """
        try:
            _ThermoRawFileAPI(path)
            return True
        except (WindowsError, IOError, ImportError):
            return False

    def infer_reader(path):
        """If the file referenced by ``path`` is a Thermo RAW
        file, return the callable (:class:`ThermoRawLoader`) to
        open it, otherwise raise an exception.

        Parameters
        ----------
        path: :class:`str`
            The path to test

        Returns
        -------
        :class:`type`:
            The type to use to open the file

        Raises
        ------
        :class:`ValueError`:
            If the file is not a Thermo RAW file
        """
        if is_thermo_raw_file(path):
            return ThermoRawLoader
        raise ValueError("Not Thermo Raw File")

    def determine_if_available():
        """Checks whether or not the COM-based Thermo
        RAW file reading feature is available.

        This is done by attempting to instantiate the
        COM-provided object, which queries the Windows
        registry for the MSFileReader.dll.

        Returns
        -------
        :class:`bool`:
            Whether or not the feature is enabled.
        """
        try:
            _ThermoRawFileAPI.create_com_object()
            return True
        except ImportError:
            return False
except ImportError as e:  # pragma: no cover
    message = str(e)

    def is_thermo_raw_file(*args, **kwargs):
        """Detect whether or not the file referenced by ``path``
        is a Thermo RAW file.

        Parameters
        ----------
        path: :class:`str`
            The path to test

        Returns
        -------
        :class:`bool`:
            Whether or not the file is a Thermo RAW file.
        """
        return False


    def infer_reader(*args, **kwargs):
        """If the file referenced by ``path`` is a Thermo RAW
        file, return the callable (:class:`ThermoRawLoader`) to
        open it, otherwise raise an exception.

        Parameters
        ----------
        path: :class:`str`
            The path to test

        Returns
        -------
        :class:`type`:
            The type to use to open the file

        Raises
        ------
        :class:`ValueError`:
            If the file is not a Thermo RAW file
        """
        raise ValueError(message)


    def register_dll(*args, **kwargs):
        """Register the location of Thermo's MSFileReader.dll
        with the COM interop system.
        """
        warnings.warn(f"no-op: {message}")
        return False


    def determine_if_available():
        """Checks whether or not the COM-based Thermo
        RAW file reading feature is available.

        This is done by attempting to instantiate the
        COM-provided object, which queries the Windows
        registry for the MSFileReader.dll.

        Returns
        -------
        :class:`bool`:
            Whether or not the feature is enabled.
        """
        warnings.warn(f"no-op: {message}")
        return False

    def _ThermoRawFileAPI(*args, **kwargs):
        raise ValueError(message)


class ThermoRawDataInterface(ScanDataSource):
    """:class:`~.ScanDataSource` implementation for Thermo's MSFileReader API.
    """
    def _scan_index(self, scan):
        return scan.scan_number - 1

    def _scan_id(self, scan):
        return _make_id(scan.scan_number)

    def _scan_time(self, scan):
        return self._source.RTFromScanNum(
            scan.scan_number)

    def _ms_level(self, scan):
        return self._source.GetMSOrderForScanNum(
            scan.scan_number)

    def _is_profile(self, scan):
        return self._source.IsProfileScanForScanNum(
            scan.scan_number)

    def _polarity(self, scan):
        filter_string = self._filter_string(scan)
        return filter_string.data['polarity']

    def _filter_string(self, scan):
        if scan.filter_string is None:
            scan.filter_string = FilterString(self._source.GetFilterForScanNum(scan.scan_number))
        return scan.filter_string

    def _scan_title(self, scan):
        return f"{self._scan_id(scan)} {self._filter_string(scan)}"

    def _scan_arrays(self, scan):
        with safearray_as_ndarray:
            arrays, _ = self._source.GetMassListFromScanNum(
                scan.scan_number)
        mz, intensity = arrays
        if isinstance(mz, np.ndarray):
            return mz.copy(), intensity.copy()
        return np.array(mz), np.array(intensity)

    def _precursor_information(self, scan):
        if self._ms_level(scan) == 1:
            return None
        scan_number = scan.scan_number
        pinfo_struct = self._source.GetPrecursorInfoFromScanNum(scan_number)
        trailer = self._trailer_values(scan)
        precursor_scan_number = trailer.get('Master Scan Number')
        if precursor_scan_number is not None:
            precursor_scan_number = int(precursor_scan_number) - 1
            if self.get_scan_by_index(precursor_scan_number).ms_level >= self._ms_level(scan):
                precursor_scan_number = None

        labels, _, _ = self._source.GetAllMSOrderData(scan_number)
        if pinfo_struct:
            # this struct field is unreliable and may fall outside the
            # isolation window
            mz = pinfo_struct.monoIsoMass
            charge = pinfo_struct.chargeState
            intensity = float(labels.intensity[0])
            # this struct field is unreliable, and simple to infer
            # precursor_scan_number = pinfo_struct.scanNumber + 1
        else:
            mz = labels.mass[0]
            intensity = float(labels.intensity[0])
            charge = labels.charge[0]
        if not charge:
            charge = ChargeNotProvided
        _mz = trailer.get('Monoisotopic M/Z', 0.0)
        # prefer the trailer m/z if available?
        if _mz > 0:
            mz = _mz

        # imitate proteowizard's firmware bug correction
        isolation_window = self._isolation_window(scan)
        if (isolation_window.upper + isolation_window.lower) / 2 <= 2.0:
            if (isolation_window.target - 3.0 > mz) or (isolation_window.target + 2.5 < mz):
                mz = isolation_window.target
        elif mz not in isolation_window:
            mz = isolation_window.target
        _charge = trailer.get('Charge State', 0)
        if _charge != 0:
            charge = _charge
        # Guess which previous scan was the precursor by iterating
        # backwards until a scan is found with a lower MS level
        if precursor_scan_number is None:
            last_index = self._scan_index(scan) - 1
            current_level = self._ms_level(scan)
            i = 0
            while last_index >= 0 and i < 100:
                prev_scan = self.get_scan_by_index(last_index)
                if prev_scan.ms_level >= current_level:
                    last_index -= 1
                else:
                    precursor_scan_number = prev_scan._data.scan_number
                    break
                i += 1
        if intensity is None:
            intensity = 0.0
        if mz is None:
            mz = 0.0
        if charge is None or charge == 0:
            charge = ChargeNotProvided
        pinfo = PrecursorInformation(
            mz, intensity, charge, _make_id(precursor_scan_number),
            source=self,
            product_scan_id=_make_id(scan.scan_number))
        return pinfo

    def _activation(self, scan):
        filter_string = self._filter_string(scan)
        tandem_sequence = filter_string.get("tandem_sequence")
        # If the tandem sequence exists, the last entry is the most recent tandem acquisition.
        # It will list contain one or more activation types. Alternatively, multiple activations
        # of the same precursor may exist in the list as separate events in the tandem sequence.
        if tandem_sequence is not None:
            activation_event = tandem_sequence[-1]
            activation_type = list(activation_event.get("activation_type"))
            has_supplemental_activation = filter_string.get("supplemental_activation")

            if activation_type is not None:
                energy = list(activation_event.get("activation_energy"))
                if len(tandem_sequence) > 1:
                    prev_event = tandem_sequence[-2]
                    # Merge previous tandem sequences of the same precursor
                    if abs(prev_event['isolation_mz'] - activation_event['isolation_mz']) < 1e-3:
                        activation_type = list(prev_event.get("activation_type")) + activation_type
                        energy = list(prev_event.get("activation_energy")) + energy
                        has_supplemental_activation = True

                if has_supplemental_activation and len(activation_type) > 1:
                    activation_type.append(supplemental_term_map[
                        dissociation_methods_map[activation_type[-1]]])
                if len(activation_type) == 1:
                    return ActivationInformation(activation_type[0], energy[0])
                else:
                    return MultipleActivationInformation(activation_type, energy)
        return None

    def _get_scan_segment(self, scan):
        trailer = self._trailer_values(scan)
        try:
            return int(trailer['Scan Segment'])
        except KeyError:
            return 1

    def _get_scan_event(self, scan):
        trailer = self._trailer_values(scan)
        try:
            return int(trailer['Scan Event'])
        except KeyError:
            return 1

    def _isolation_window(self, scan):
        ms_level = self._ms_level(scan)
        if ms_level == 1:
            return None
        isolation_width = 0
        trailer = self._trailer_values(scan)
        try:
            isolation_width = trailer[f'MS{ms_level} Isolation Width']
        except KeyError:
            segment = self._get_scan_segment(scan)
            event = self._get_scan_event(scan)
            isolation_width = self._method.isolation_width_for(segment, event=event)
            if not isolation_width:
                isolation_width = self._method.isolation_width_for(segment, ms_level=ms_level)
        if isolation_width == 0:
            return None
        isolation_width /= 2.
        isolation_mz = self._source.GetPrecursorMassForScanNum(scan.scan_number, ms_level)
        return IsolationWindow(isolation_width, isolation_mz, isolation_width)

    def _instrument_configuration(self, scan):
        fline = self._filter_string(scan)
        try:
            confid = self._analyzer_to_configuration_index[analyzer_map[fline.data.get("analyzer")]]
            return self._instrument_config[confid]
        except KeyError:
            return None

    def _trailer_values(self, scan):
        if scan.trailer_values is not None:
            return scan.trailer_values
        trailer_extras = self._source.GetTrailerExtraForScanNum(scan.scan_number)
        scan.trailer_values = trailer_extras
        return trailer_extras

    def _acquisition_information(self, scan):
        fline = self._filter_string(scan)
        event = self._get_scan_event(scan)
        trailer_extras = self._trailer_values(scan)
        traits = {
            'preset scan configuration': event,
            'filter string': fline,
        }
        cv = fline.get("compensation_voltage")
        if cv is not None:
            traits[FAIMS_compensation_voltage] = cv
        event = ScanEventInformation(
            self._scan_time(scan),
            injection_time=unitfloat(trailer_extras.get(
                'Ion Injection Time (ms)', 0.0), 'millisecond'),
            window_list=[ScanWindow(
                fline.get("scan_window")[0], fline.get("scan_window")[1])],
            traits=traits)
        return ScanAcquisitionInformation("no combination", [event])

    def _annotations(self, scan):
        fline = self._filter_string(scan)
        trailer_extras = self._trailer_values(scan)
        annots = {
            "filter string": fline,
        }
        microscans = trailer_extras.get("Micro Scan Count")
        if microscans is not None:
            annots['[Thermo Trailer Extra]Micro Scan Count'] = microscans
        scan_segment = trailer_extras.get("Scan Segment")
        if scan_segment is not None:
            annots['[Thermo Trailer Extra]Scan Segment'] = scan_segment
        scan_event = trailer_extras.get("Scan Event")
        if scan_event is not None:
            annots['[Thermo Trailer Extra]Scan Event'] = scan_event
        mono_mz = trailer_extras.get("Monoisotopic M/Z")
        if mono_mz is not None and mono_mz > 0:
            annots['[Thermo Trailer Extra]Monoisotopic M/Z'] = mono_mz
        hcd_ev = trailer_extras.get('HCD Energy eV')
        if hcd_ev is not None and hcd_ev > 0:
            annots['[Thermo Trailer Extra]HCD Energy eV'] = hcd_ev
        return annots


class ThermoRawLoader(ThermoRawDataInterface, RandomAccessScanSource, _RawFileMetadataLoader):
    """Reads scans from Thermo Fisher RAW files directly. Provides both iterative and
    random access.
    """

    def __init__(self, source_file, _load_metadata=True, **kwargs):
        self.source_file = source_file
        self._source = _ThermoRawFileAPI(self.source_file)
        warnings.warn(("ThermoRawLoader COM implementation is deprecated."
                       " Please consider switching to the newer .NET implementation"), DeprecationWarning)
        self._producer = None
        self._scan_type_index = dict()
        self.make_iterator()
        self.initialize_scan_cache()
        self._index = self._pack_index()
        self._first_scan_time = self.get_scan_by_index(0).scan_time
        self._last_scan_time = self.get_scan_by_id(self._source.LastSpectrumNumber).scan_time
        if _load_metadata:
            self._method = self._parse_method()
            self._build_scan_type_index()
            self._get_instrument_info()

    def _get_instrument_model_name(self):
        return self._source.GetInstModel()

    def _get_instrument_serial_number(self):
        return self._source.GetInstSerialNumber()

    def samples(self):
        """Describe the sample(s) used to generate the mass spectrometry
        data contained in this file.

        Returns
        -------
        :class:`list` of :class:`~.Sample`
        """
        result = []
        si = self._source
        raw_name = os.path.basename(self.source_file_name)
        sample = Sample(si.GetSeqRowSampleID() or raw_name)
        sample.name = si.GetSeqRowSampleName() or si.GetSeqRowSampleID()
        if si.GetSeqRowSampleVolume():
            sample.parameters['sample volume'] = si.GetSeqRowSampleVolume()
        if si.GetSeqRowSampleWeight():
            sample.parameters['sample mass'] = si.GetSeqRowSampleWeight()
        if si.GetSeqRowVial():
            sample.parameters['sample vial'] = si.GetSeqRowVial()
        if si.GetSeqRowBarcode():
            sample.parameters['sample barcode'] = si.GetSeqRowBarcode()
        result.append(sample)
        return result

    def _parse_method(self):
        return _InstrumentMethod(self._source.GetInstMethod())

    def _has_ms1_scans(self):
        if self._scan_type_index:
            return 1 in self._scan_type_index
        else:
            # metadata has not been loaded so best to assume there is
            return True

    def _has_msn_scans(self):
        if self._scan_type_index:
            return max(self._scan_type_index) > 1
        else:
            # metadata has not been loaded so best to assume there is
            return True

    def has_msn_scans(self):
        return self._has_msn_scans()

    def has_ms1_scans(self):
        return self._has_ms1_scans()

    def __reduce__(self):
        return self.__class__, (self.source_file, False), self.__getstate__()

    def __getstate__(self):
        state = {
            "method": self._method,
            "scan_type_index": self._scan_type_index,
            "analyzer_to_configuration_index": self._analyzer_to_configuration_index,
            "instrument_config": self._instrument_config,
            "previous_ms_levels": self._previous_ms_levels,
        }
        return state

    def __setstate__(self, state):
        self._method = state['method']
        self._scan_type_index = state['scan_type_index']
        self._analyzer_to_configuration_index = state['analyzer_to_configuration_index']
        self._instrument_config = state['instrument_config']
        self._previous_ms_levels = state['previous_ms_levels']

    @property
    def index(self):
        """Accesses the scan index

        Returns
        -------
        :class:`collections.OrderedDict`
            Maps scan ID to index
        """
        return self._index

    def __repr__(self):
        return f"ThermoRawLoader({self.source_file})"

    def _pack_index(self):
        index = OrderedDict()
        for sn in range(1, self._source.NumSpectra + 1):
            index[_make_id(sn)] = sn
        return index

    def __len__(self):
        return len(self.index)

    def close(self):
        """Close the underlying file reader.
        """
        if self._source is not None:
            self._source.Close()
            self._source = None
        self._dispose()

    def __del__(self):
        self.close()

    def reset(self):
        self.make_iterator(None)
        self.initialize_scan_cache()

    def _scan_time_to_scan_number(self, scan_time):
        if scan_time < self._first_scan_time:
            scan_time = self._first_scan_time
        elif scan_time > self._last_scan_time:
            scan_time = self._last_scan_time
        scan_number = self._source.ScanNumFromRT(scan_time)
        return scan_number

    def get_scan_by_id(self, scan_id):
        """Retrieve the scan object for the specified scan id.

        If the scan object is still bound and in memory somewhere,
        a reference to that same object will be returned. Otherwise,
        a new object will be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        scan_number = int(str(scan_id).replace(_id_template, ''))
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            if not package.validate(self):
                raise KeyError(str(scan_id)) from None
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def get_scan_by_index(self, index):
        """Retrieve the scan object for the specified scan index.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        index: int
            The index to get the scan for

        Returns
        -------
        Scan
        """
        scan_number = int(index) + 1
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            if not package.validate(self):
                raise IndexError(index) from None
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def get_scan_by_time(self, time):
        """Retrieve the scan object for the specified scan time.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        time : float
            The time to get the nearest scan from

        Returns
        -------
        Scan
        """
        if time < self._first_scan_time:
            time = self._first_scan_time
        elif time > self._last_scan_time:
            time = self._last_scan_time
        scan_number = self._scan_time_to_scan_number(time)
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True, **kwargs):
        """Reconstruct an iterator which will start from the scan matching one of ``scan_id``,
        ``rt``, or ``index``. Only one may be provided.

        After invoking this method, the iterator this object wraps will be changed to begin
        yielding scan bunchs (or single scans if ``grouped`` is ``False``).

        Arguments
        ---------
        scan_id: str, optional
            Start from the scan with the specified id.
        rt: float, optional
            Start from the scan nearest to specified time (in minutes) in the run. If no
            exact match is found, the nearest scan time will be found, rounded up.
        index: int, optional
            Start from the scan with the specified index.
        require_ms1: bool, optional
            Whether the iterator must start from an MS1 scan. True by default.
        grouped: bool, optional
            whether the iterator should yield scan bunches or single scans. True by default.
        """
        if scan_id is not None:
            scan_number = int(str(scan_id).replace(_id_template, '')) - 1
        elif index is not None:
            scan_number = int(index)
        elif rt is not None:
            scan_number = self._scan_time_to_scan_number(rt)
        if require_ms1:
            start_index = scan_number
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            scan_number = start_index
        iterator = self._make_pointer_iterator(start_index=scan_number)
        if grouped:
            self._producer = self._scan_group_iterator(iterator, grouped, **kwargs)
        else:
            self._producer = self._single_scan_iterator(iterator, grouped, **kwargs)
        return self

    def _make_scan_index_producer(self, start_index=None, start_time=None):
        if start_index is not None:
            return range(start_index + 1, self._source.NumSpectra + 1)
        elif start_time is not None:
            start_index = self._scan_time_to_scan_number(start_time)
            while start_index != 1:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            return range(start_index, self._source.NumSpectra + 1)
        else:
            return range(1, self._source.NumSpectra + 1)

    def _make_pointer_iterator(self, start_index=None, start_time=None):
        iterator = self._make_scan_index_producer(start_index, start_time)
        for i in iterator:
            yield ThermoRawScanPtr(i)

    def _make_default_iterator(self):
        return self._make_pointer_iterator()

    def _make_cache_key(self, scan):
        return scan._data.scan_number

    def next(self):
        return next(self._producer)
